/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvcontour.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qvdta/qvcontour.h>

namespace qvdta
{
// Direction-number		Y
//	NE-7	N-0	NW-1	|
//	E-6	*	W-2	v
//	SE-5	S-4	SW-3
// X -->
//						N	NO	O	SO	S	SE	E	NE
/// \cond
const char	coorX8Connect[8] =	{	0,	1,	1,	1,	0,	-1,	-1,	-1	};
const char	coorY8Connect[8] =	{	-1,	-1,	0,	1,	1,	1,	0,	-1	};
const char	coorX4Connect[4] =	{	0,		1,		0,		-1,		};
const char	coorY4Connect[4] =	{	-1,		0,		1,		0,		};
const char	coorX4Diag[8] =		{		1,		1,		-1,		-1	};
const char	coorY4Diag[8] =		{		-1,		1,		1,		-1	};
/// \endcond

// Auxiliar function for border extraction. It gets a border point, and the direction where there is one of the outside of the connected-set pixels.
/// \cond
QVPolyline getConnectedSetBorderContourThresholdFromBorderPoint(const QVImage<uChar> &image, const int startPointX, const int startPointY, const uChar threshold)
	{
	QVPolyline lista;

	lista.closed = true;
	lista.append(QPoint(startPointX, startPointY));

	QVIMAGE_INIT_READ(uChar,image);
	QRect roi = image.getROI();

	Q_ASSERT_X(roi.contains(startPointX, startPointY), "getContourThresholdFromBorderPoint", "start point out of image ROI");
	Q_ASSERT_X(QVIMAGE_PIXEL(image, startPointX, startPointY, 0) >= threshold, "getContourThresholdFromBorderPoint", "start point is not contained in a connected set");

	// We check this is not an interior pixel, neither a solitary one.
	// Also we look for a neighbour pixel not belonging to any connected set.
	uChar searchDir = 128, numOuterPixels = 0;
	for (int i = 0; i<8; i++)
		{
		int x =  startPointX +coorX8Connect[i], y =  startPointY +coorY8Connect[i];
		if (!roi.contains(x, y))
			{
			numOuterPixels++;
			searchDir = i;
			}
		else if (QVIMAGE_PIXEL(image, x, y,0) < threshold)
			{
			numOuterPixels++;
			searchDir = i;
			}
		}

	// Case we receive an interior pixel, raise assert.
	Q_ASSERT_X(searchDir < 8, "getContourThresholdFromBorderPoint", "start point is inside the set, not in the border");

	// Case we have a solitary pixel, we return that pixel.
	if (numOuterPixels == 8)
		return lista;

	// We insert each point of the border contour, inserting it to the point list.
	int sumSearchDir = 0, actualPointX = startPointX, actualPointY = startPointY;
	while (true)
		{
		// We search for the next point belonging to the contour.
		uChar d;
		int	nextPointX, nextPointY;
		for (d = 0; d < 8; d++)
			{
			searchDir = (searchDir+1)%8;
			nextPointX = actualPointX + coorX8Connect[searchDir];
			nextPointY = actualPointY + coorY8Connect[searchDir];
			if (roi.contains(nextPointX, nextPointY))
				if ( (QVIMAGE_PIXEL(image, nextPointX, nextPointY,0) >= threshold) )
					break;
			}

		sumSearchDir += d - 3;

		actualPointX = nextPointX;
		actualPointY = nextPointY;

		if ( QVIMAGE_PIXEL(image, actualPointX, actualPointY,0) < threshold )
			break;

		if ( startPointX == actualPointX && startPointY == actualPointY)
			break;

		lista.append(QPoint(actualPointX, actualPointY));
		searchDir = searchDir + 4;
		}

	lista.direction = (sumSearchDir >= 0);
	return lista;
	}
/// \endcond

QVPolyline getConnectedSetBorderContourThreshold(const QVImage<uChar> &image, const QPoint startPoint, const uChar threshold)
	{
	QVIMAGE_INIT_READ(uChar,image);
	const QRect roi = image.getROI();

	int col = startPoint.x(), row = startPoint.y();

	if (QVIMAGE_PIXEL(image, col, row,0) < threshold)
		return QVPolyline();

	while (roi.contains(col+1, row))
		{
		if ( QVIMAGE_PIXEL(image, col+1, row,0) < threshold )
			break;
		col++;
		}

	return getConnectedSetBorderContourThresholdFromBorderPoint(image, col, row, threshold);
	}

QList<QVPolyline> getConnectedSetBorderContoursThreshold(const QVImage <uChar> &image, const uChar threshold)
	{
	qDebug() << "getPolylinesThreshold()";
	QVImage<uChar> mask(image.getCols()+1, image.getRows()+1);
	qvipp::Set(mask,0);

	QVIMAGE_INIT_READ(uChar,image);
	QVIMAGE_INIT_WRITE(uChar,mask);

	QRect roi = image.getROI();

	QList<QVPolyline> polylineList;

	// We look for pixels contained in a connected set (gray-level value >= threshold) in the image
	for (int row = roi.y(); row < roi.y() + roi.height(); row++)
		for (int col = roi.x(); col < roi.y() + roi.width(); col++)
			{
			// If we find any pixel like that, we can be sure (because the search we did) it belongs to it's border.
			if (QVIMAGE_PIXEL(image, col, row,0) >= threshold)
				{
				// if pixel is not marked, we get it's contour
				if ( !QVIMAGE_PIXEL(mask, col, row,0) )
					{
					QVPolyline lista = getConnectedSetBorderContourThresholdFromBorderPoint(image, col, row, threshold);
					draw(mask, lista, true, true);
					polylineList.append(lista);
					}

				// We ensure next pixel we process will not belong to a connected set.
				while (roi.contains(col+1, row))
					{
					if ( QVIMAGE_PIXEL(image, col+1, row,0) < threshold )
						break;
					col++;
					}

				// This is for the case in which we find an internal contour, that has not been processed and marked.
				if ( !QVIMAGE_PIXEL(mask, col, row,0) )
					{
					QVPolyline lista = getConnectedSetBorderContourThresholdFromBorderPoint(image, col, row, threshold);
					draw(mask, lista, true, true);
					polylineList.append(lista);
					}
				}

			}
	qDebug() << "getPolylinesThreshold():"<< polylineList.size() << "contours obtained";
	qDebug() << "getPolylinesThreshold() <~ return";
	return polylineList;
	}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

QVPolyline getLineContourThreshold4Connectivity(QVImage<uChar> &image, const QPoint point, QVPolyline &polyline, const uChar threshold, bool reverse)
	{
	const uInt cols = image.getCols(), rows = image.getRows();
	QVIMAGE_INIT_WRITE(uChar, image);

	uInt lastDir = 666, coorX = point.x(), coorY = point.y();

	qDebug() << "\tContour: new contour";

	while(true)
		{
		qDebug() << "\tContour:\tAppending point (" << coorX << ", " << coorY << ")";
		if (reverse)
			polyline.prepend(QPoint(coorX, coorY));
		else
			polyline.append(QPoint(coorX, coorY));

		QVIMAGE_PIXEL(image, coorX, coorY, 0) = 0;

		uInt dir;
		int newCoorX, newCoorY;
		for (dir = 0; dir < 4; dir++)
			{
			newCoorX = coorX + coorX4Connect[dir];
			newCoorY = coorY + coorY4Connect[dir];

			// Check if we are inside the limits in that direction
			if ( (newCoorX < 0) || (newCoorY < 0) || (newCoorX >= cols) || (newCoorY >= rows) )
				continue;

			// Check if it is a valid direction and if the pixel in that direction is part of a contour
			if ( (QVIMAGE_PIXEL(image, newCoorX, newCoorY, 0) >= threshold)  && (lastDir != dir) )
				break;
			}

		if (dir == 4) break;

		coorX = newCoorX;
		coorY = newCoorY;
		lastDir = (dir+2)%4;
		}

	return polyline;
	}

QList<QVPolyline> getLineContoursThreshold4Connectivity(const QVImage<uChar> &image, const uChar threshold)
	{
	const uInt cols = image.getCols(), rows = image.getRows();
	QVImage<uChar> clone = image;

	QList<QVPolyline> polylineList;

	// Transverse the image
	for(uInt col = 0; col < cols; col++)
		for(uInt row = 0; row < rows; row++)
			{
			QVIMAGE_INIT_READ(uChar, clone);
			// If we don't have an active pixel, continue
			if ( (QVIMAGE_PIXEL(clone, col, row, 0) < threshold) )
				continue;

			// Else, we compose the contour following two active neighbour pixels:
			QVPolyline polyline;

			// We follow first active neighbour pixel, composing the list of pixels in direct order
			getLineContourThreshold4Connectivity(clone, QPoint(col, row), polyline, threshold, false);

			// Find another neighbour close to the pixel.
			uInt dir;			
			int	newCoorX, newCoorY;
			for (dir = 0; dir < 4; dir++)
				{
				newCoorX = col + coorX4Connect[dir];
				newCoorY = row + coorY4Connect[dir];
	
				// Check if we are inside the limits in that direction
				if ( (newCoorX < 0) || (newCoorY < 0) || (newCoorX >= cols) || (newCoorY >= rows) )
					continue;
	
				// Check if it is a valid direction and if the pixel in that direction is part of a contour
				if ( (clone(newCoorX, newCoorY) >= threshold) )
					break;
				}

			// If we found it, add the contour in reverse order.
			if (dir != 4)
				getLineContourThreshold4Connectivity(clone, QPoint(newCoorX, newCoorY), polyline, threshold, true);

			// Finally add the polyline to the list.
			polylineList.append(polyline);
			}

	return polylineList;
	}

////////////////////////////////////////////////////////////////////////////////////
// Replicated functions from 4-connected version.
QVPolyline getLineContourThreshold8Connectivity(QVImage<uChar> &image, const QPoint point, QVPolyline &polyline, const uChar threshold, bool reverse)
	{
	const uInt cols = image.getCols(), rows = image.getRows();
	QVIMAGE_INIT_WRITE(uChar, image);

	uInt lastDir = 666, coorX = point.x(), coorY = point.y();

	qDebug() << "\tContour: new contour";

	bool continueCond = true;
	while(continueCond)
		{
		qDebug() << "\tContour:\tAppending point (" << coorX << ", " << coorY << ")";
		if (reverse)
			polyline.prepend(QPoint(coorX, coorY));
		else
			polyline.append(QPoint(coorX, coorY));

		QVIMAGE_PIXEL(image, coorX, coorY, 0) = 0;

		// Buscamos un píxel en los vecinos 4 conectados.
		uInt dir;
		int newCoorX, newCoorY;
		for (dir = 0; dir < 4; dir++)
			{
			newCoorX = coorX + coorX4Connect[dir];
			newCoorY = coorY + coorY4Connect[dir];

			// Check if we are inside the limits in that direction
			if ( (newCoorX < 0) || (newCoorY < 0) || (newCoorX >= cols) || (newCoorY >= rows) )
				continue;

			// Check if it is a valid direction and if the pixel in that direction is part of a contour
			if ( (QVIMAGE_PIXEL(image, newCoorX, newCoorY, 0) >= threshold)  && (lastDir != dir) )
				break;
			}

		if (dir == 4) 
			{
			// Buscamos un píxel en los vecinos 4 conectados diagonalmente.
			uInt dir;
			int newCoorX, newCoorY;
			for (dir = 0; dir < 4; dir++)
				{
				newCoorX = coorX + coorX4Diag[dir];
				newCoorY = coorY + coorY4Diag[dir];
	
				// Check if we are inside the limits in that direction
				if ( (newCoorX < 0) || (newCoorY < 0) || (newCoorX >= cols) || (newCoorY >= rows) )
					continue;
	
				// Check if it is a valid direction and if the pixel in that direction is part of a contour
				if ( (QVIMAGE_PIXEL(image, newCoorX, newCoorY, 0) >= threshold)  && (lastDir != dir) )
					break;
				}
			if (dir == 4) break;

			coorX = newCoorX;
			coorY = newCoorY;
			lastDir = (dir+2)%4;
			}
		else	{
			coorX = newCoorX;
			coorY = newCoorY;
			lastDir = (dir+2)%4;
			}
		}

	return polyline;
	}

QList<QVPolyline> getLineContoursThreshold8Connectivity(const QVImage<uChar> &image, const uChar threshold)
	{
	const uInt cols = image.getCols(), rows = image.getRows();
	QVImage<uChar> clone = image;

	QList<QVPolyline> polylineList;

	// Transverse the image
	for(uInt col = 0; col < cols; col++)
		for(uInt row = 0; row < rows; row++)
			{
			QVIMAGE_INIT_READ(uChar, clone);
			// If we don't have an active pixel, continue
			if ( (QVIMAGE_PIXEL(clone, col, row, 0) < threshold) )
				continue;

			// Else, we compose the contour following two active neighbour pixels:
			QVPolyline polyline;

			// We follow first active neighbour pixel, composing the list of pixels in direct order
			getLineContourThreshold8Connectivity(clone, QPoint(col, row), polyline, threshold, false);

			// Find another neighbour close to the pixel, in 4 connected neighbours
			uInt dir;			
			int	newCoorX, newCoorY;
			for (dir = 0; dir < 4; dir++)
				{
				newCoorX = col + coorX4Connect[dir];
				newCoorY = row + coorY4Connect[dir];
	
				// Check if we are inside the limits in that direction
				if ( (newCoorX < 0) || (newCoorY < 0) || (newCoorX >= cols) || (newCoorY >= rows) )
					continue;
	
				// Check if it is a valid direction and if the pixel in that direction is part of a contour
				if ( (clone(newCoorX, newCoorY) >= threshold) )
					break;
				}

			// If we found it, add the contour in reverse order.
			if (dir != 4)
				getLineContourThreshold8Connectivity(clone, QPoint(newCoorX, newCoorY), polyline, threshold, true);
			else	{
				// Find another neighbour close to the pixel, in diagonal connected neighbours
				uInt dir;			
				int	newCoorX, newCoorY;
				for (dir = 0; dir < 4; dir++)
					{
					newCoorX = col + coorX4Diag[dir];
					newCoorY = row + coorY4Diag[dir];
		
					// Check if we are inside the limits in that direction
					if ( (newCoorX < 0) || (newCoorY < 0) || (newCoorX >= cols) || (newCoorY >= rows) )
						continue;
		
					// Check if it is a valid direction and if the pixel in that direction is part of a contour
					if ( (clone(newCoorX, newCoorY) >= threshold) )
						break;
					}
	
				// If we found it, add the contour in reverse order.
				if (dir != 4)
					getLineContourThreshold8Connectivity(clone, QPoint(newCoorX, newCoorY), polyline, threshold, true);
				}

			// Finally add the polyline to the list.
			polylineList.append(polyline);
			}

	return polylineList;
	}

}
