/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvpolyline.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qvdta/qvpolyline.h>

#define CONST_PI	((double)3.14159)
#define SIGN(X)		(((X)>=0)?1:-1)

namespace qvdta
{
//////////////////////////////////////////////
// QVPolylines

// IPE
#define xA(Index) (coorX[previous[Index]])
#define yA(Index) (coorY[previous[Index]])
#define xB(Index) (coorX[Index])
#define yB(Index) (coorY[Index])
#define xC(Index) (coorX[next[Index]])
#define yC(Index) (coorY[next[Index]])
#define	ABS(X)	(((X)>=0)?(X):-(X))
#define TRIANGLE_AREA(I) ABS(xA(I)*yC(I) - xA(I)*yB(I) + xB(I)*yA(I) - xB(I)*yC(I) + xC(I)*yB(I) - xC(I)*yA(I))

void IterativePointElimination(const QVPolyline &polyline, QVPolyline &result, const double minAreaDiv2)
	{
	// Actually, TRIANGLE_AREA does computes the double of the area of the triangle. We double the parameter value
	// to correctly adjust it.
	const uInt polylineSize = polyline.size(), minArea = minAreaDiv2 * 2;

	if (polylineSize < 3) return;

	// We create a static linked list of points with arrays previous, next, coorX, coorY,
	// from the polyline.	
	uInt	*previous = new uInt[polylineSize],
		*next = new uInt[polylineSize],
		*valid = new uInt[polylineSize],
		*area = new uInt[polylineSize],
		discartablePoints = 0,
		validPoints = polylineSize;

	int	*coorX = new int[polylineSize],
		*coorY = new int[polylineSize];

	for (uInt i=0; i< polylineSize; i++)
		{
		previous[i] = i-1;
		next[i] = i+1;
		coorX[i] = polyline.at(i).x();
		coorY[i] = polyline.at(i).y();
		}

	previous[0] = polylineSize-1;
	next[polylineSize-1] = 0;

	// Create valid index and area initialization
	for (uInt i=0; i < polylineSize; i++)
		{
		valid[i] = true;
		area[i] = TRIANGLE_AREA(i);
		}

	// This is where points are actually deleted.
	// Well, not deleted, but are marked as not valid,
	// and latter when adding points to result
	// QVPolyline, not valid points will not be added.
	uInt index = 0, lastIndex = previous[index];
	bool cond = true;

	while(cond && validPoints > 3)
		{
		Q_ASSERT(valid[index]);
		Q_ASSERT(valid[next[index]]);
		Q_ASSERT(valid[previous[index]]);
		if (area[index] < minArea)
			{
			// we eliminate the point from the set
			valid[index] = false;
			next[previous[index]] = next[index];
			previous[next[index]] = previous[index];
			lastIndex = previous[index];
			validPoints--;

			// step and actualize contiguous points areas
			index = next[index];
			area[index] = TRIANGLE_AREA(index);
			area[previous[index]] = TRIANGLE_AREA(previous[index]);
			}
		else	{
			// end condition: a whole turn around the linked list with no
			// eliminations.
			if (index == lastIndex)
				cond = false;

			// step
			index = next[index];
			}
		}

	// After IPE, if we have a contour of 3 points, return if the triangle
	// has less area than minimum.
	if (validPoints == 3)
		{
		Q_ASSERT(valid[index]);
		Q_ASSERT(area[index] == area[next[index]]);
		Q_ASSERT(area[index] == area[previous[index]]);
		Q_ASSERT(index == previous[previous[index]]);

		if (area[index] < minArea)
			return;
		}

	// Store the resulting points in result polyline.
	for (uInt n=0; n< polylineSize; n++)
		if (valid[n])
			result.append(polyline.at(n));
	result.closed = polyline.closed;
	result.direction = polyline.direction;

	delete previous;
	delete next;
	delete valid;
	delete area;
	delete coorX;
	delete coorY;
	}

// Draw
QVPolyline::QVPolyline(): QList<QPoint>(),
	closed(false), direction(false)//, dir(0)
	{
	qDebug() << "QVPolyline()";
	qDebug() << "QVPolyline() <~ return";
	};

QVPolyline::QVPolyline(const QVPolyline &polyline): QList<QPoint>(polyline),
	closed(polyline.closed), direction(polyline.direction)//, dir(polyline.dir)
	{
	qDebug() << "QVPolyline(const QVPolyline &)";
	qDebug() << "QVPolyline(const QVPolyline &) <~ return";
	};

QPoint	linesIntersection(QPoint a, QPoint b, QPoint c, QPoint d)
	// gets the intersection point for lines ab and cd
	{
	double x1 = a.rx(), x2 = b.rx(), x3 = c.rx(), x4 = d.rx();
	double y1 = a.ry(), y2 = b.ry(), y3 = c.ry(), y4 = d.ry();

	double denominador = (y4 - y3)*(x2 - x1) - (x4 - x3)*(y2 - y1);
	if (denominador == 0)
		return QPoint( (int)(b.rx() + c.rx())/2, (int)(b.ry() + c.ry())/2 );

	double	ua = (x4 - x3)*(y1 - y3) - (y4 - y3)*(x1 - x3),
		ub = (x2 - x1)*(y1 - y3) - (y2 - y1)*(x1 - x3);

	QPoint p = QPoint(
			(int) (x1 + ua*(x2 - x1) / denominador),
			(int) (y1 + ub*(y2 - y1) / denominador)
			);
	return p;
	}

////////////////

void drawPoints(const QList<QPoint> &hotPoints, QVImage<uChar,3> &dest)
	{
	QListIterator<QPoint> iterator(hotPoints);
	uChar pixel[3] = { 255,0,0 };

	while(iterator.hasNext())
		{
		QPoint p = iterator.next();
		QVPolyline rectangle = QVPolyline::rectangle(p.x()-2, p.y()-2, p.x()+2, p.y()+2);
		draw(dest, rectangle, pixel,true,true);
		}
	}

void drawPoints(const QList<QPoint> &hotPoints, QVImage<uChar> &dest)
	{
	QListIterator<QPoint> iterator(hotPoints);
	while(iterator.hasNext())
		{
		QPoint p = iterator.next();
		QVPolyline rectangle = QVPolyline::rectangle(p.x()-2, p.y()-2, p.x()+2, p.y()+2);
		draw(dest, rectangle, 255,true,true);
		}
	}


void draw(QVImage<uChar> &image, const QVPolyline &polyline, const uChar constant, bool linked, bool safe)
	{
	Q_ASSERT_X(polyline.size() > 0, "drawPolyline()", "polyline size equals zero");

	QListIterator<QPoint> iterator(polyline);
	QRect roi = image.getROI();

	QPoint previous;

	if (linked)
		previous = iterator.next();

	QVIMAGE_INIT_WRITE(uChar,image);
	while (iterator.hasNext())
		{
		QPoint actual = iterator.next();
		int x = actual.x(), y = actual.y();
		if (!linked && !safe)
			QVIMAGE_PIXEL(image, x, y,0) = constant;
		else if (linked)
			{
			QVPolyline line = QVPolyline::line(x, y, previous.rx(), previous.ry());
			draw(image, line, constant, false, safe);
			previous = actual;
			}
		else if (roi.contains(x, y))
			QVIMAGE_PIXEL(image, x, y,0) = constant;
		}
	}

void draw(QVImage<uChar,3> &image, const QVPolyline &polyline, const uChar constant[3], bool linked, bool safe)
	{
	Q_ASSERT_X(polyline.size() > 0, "drawPolyline()", "polyline size equals zero");
	//if (polyline.size() == 0) return;

	QListIterator<QPoint> iterator(polyline);
	QRect roi = image.getROI();

	QPoint previous;

	if (linked)
		previous = iterator.next();

	QVIMAGE_INIT_WRITE(uChar,image);
	while (iterator.hasNext())
		{
		QPoint actual = iterator.next();
		int x = actual.x(), y = actual.y();
		if (!linked && !safe)
			{
			QVIMAGE_PIXEL(image, x, y,0) = constant[0];
			QVIMAGE_PIXEL(image, x, y,1) = constant[1];
			QVIMAGE_PIXEL(image, x, y,2) = constant[2];
			//image(actual.rx(), actual.ry(),0) = constant[0];
			//image(actual.rx(), actual.ry(),1) = constant[1];
			//image(actual.rx(), actual.ry(),2) = constant[2];
			}
		else if (linked)
			{
			QVPolyline line = QVPolyline::line(actual.rx(), actual.ry(), previous.rx(), previous.ry());
			draw(image, line, constant, false, safe);
			previous = actual;
			}
		else if (roi.contains(actual.rx(), actual.ry()))
			{
			QVIMAGE_PIXEL(image, x, y,0) = constant[0];
			QVIMAGE_PIXEL(image, x, y,1) = constant[1];
			QVIMAGE_PIXEL(image, x, y,2) = constant[2];
			/*image(actual.rx(), actual.ry(),0) = constant[0];
			image(actual.rx(), actual.ry(),1) = constant[1];
			image(actual.rx(), actual.ry(),2) = constant[2];*/
			}
		}
	}


void draw(QVImage<uChar> &img, const QList< QVPolyline > &polylineList, const uChar constant, bool linked, bool safe)
	{
	QListIterator<QVPolyline> iterator(polylineList);
	while (iterator.hasNext())
		draw(img, iterator.next(), constant, linked, safe);
	}

void draw(QVImage<uChar,3> &img,const QList< QVPolyline > &polylineList, const uChar constant[3],bool linked,bool safe)
	{
	QListIterator<QVPolyline> iterator(polylineList);
	while (iterator.hasNext())
		draw(img, iterator.next(), constant, linked, safe);
	}


QVPolyline QVPolyline::ellipse(uInt n, float x, float y, float maxRadio, float minRadio, float ang)
	{
	QVPolyline ellipse;
	float w = 2*CONST_PI/(n-1);
	float maxArg = (maxRadio+minRadio)/2;
	float minArg = (maxRadio-minRadio)/2;

	for (uInt t = 0; t < n; t++)
		ellipse.append(QPoint (	(uInt) (x + maxArg*cos(ang+w*t) + minArg*cos(ang-w*t)),
						(uInt) (y + maxArg*sin(ang+w*t) + minArg*sin(ang-w*t))
						));
	return ellipse;
	}

QVPolyline QVPolyline::rectangle(int x1, int y1, int x2, int y2)
	{
	QVPolyline rectangle;
	rectangle.append(QPoint( x1, y1 ));
	rectangle.append(QPoint( x1, y2 ));
	rectangle.append(QPoint( x2, y2 ));
	rectangle.append(QPoint( x2, y1 ));
	rectangle.append(QPoint( x1, y1 ));
	return rectangle;
	}

QVPolyline QVPolyline::line(int x1, int y1, int x2, int y2)
	{
	QVPolyline line;

	line.append(QPoint( x1, y1 ));

	if (x1 == x2 && y1 == y2)
		return line;

	int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;
	
	dx=x2-x1;      // the horizontal distance of the line
	dy=y2-y1;      // the vertical distance of the line
	dxabs=abs(dx);
	dyabs=abs(dy);
	sdx=SIGN(dx);
	sdy=SIGN(dy);
	x=dyabs>>1;
	y=dxabs>>1;
	px=x1;
	py=y1;
	
	if (dxabs>=dyabs) // the line is more horizontal than vertical
		for(i=0;i<dxabs;i++)
			{
			y+=dyabs;
			if (y>=dxabs)
				{
				y-=dxabs;
				py+=sdy;
				}
			px+=sdx;
			line.append(QPoint( px, py ));
			}
	else	// the line is more vertical than horizontal
		for(i=0;i<dyabs;i++)
			{
			x+=dxabs;
			if (x>=dyabs)
				{
				x-=dyabs;
				px+=sdx;
				}
			py+=sdy;
			line.append(QPoint( px, py ));
			}

	return line;
	}

/*void joinExtremes(QVPolyline &main, QVPolyline &second)
	{
	QPoint a1 = main.takeLast(), a2 = main.last(), b1 = second.at(0), b2 = second.at(1);

	main.append(linesIntersection(a1,a2, b1,b2));

	for (int i = 1; i< second.size(); i++)
		main.append(second.at(i));
	}

double dist2(QPoint p1, QPoint p2)
	{ return (p1.x() - p2.x())*(p1.x() - p2.x()) + (p1.y() - p2.y())*(p1.y() - p2.y()); }

template <typename T>QList<T> & reverse(QList<T> &p)
	{
	for (int i = 0; i< p.size()/2; i++)
		p.swap(i, p.size()-i-1);
	return p;
	}

QList<QVPolyline> * compactPolylines(const QList<QVPolyline> &polylineList, double maxDist)
	{
	//qDebug() << "compactPolylines()";
	uInt n = polylineList.size();
	//qDebug() << "compactPolylines(): size"<<n;
	//if (n < 2) return new QList<QVPolyline>(polylineList);

	bool * segmentUsed = new bool[n];
	QList<QVPolyline> * result = new QList<QVPolyline>();

	for (uInt i=0; i<n; i++)
		segmentUsed[i] = false;

	for (uInt i=0; i<n-1; i++)
		{
		//qDebug() << "compactPolylines(): segmento" << i << segmentUsed[i];
		if (segmentUsed[i]) continue;
		QVPolyline actual = polylineList.at(i);
		if (actual.size() < 2) continue;

		for (uInt j= i+1; j<n; j++)
			{
			if (segmentUsed[j]) continue;
			QVPolyline next = polylineList.at(j);
			if (next.size() < 2) continue;

			if (dist2(actual.back(), next.front()) < maxDist)
				{ append(actual, next); segmentUsed[j] = true; }

			else if (dist2(actual.back(), next.back()) < maxDist)
				{ reverse(next); append(actual, next); segmentUsed[j] = true; }

			else if (dist2(actual.front(), next.front()) < maxDist)
				{ reverse(actual); append(actual, next); segmentUsed[j] = true; }

			else if (dist2(actual.front(), next.back()) < maxDist)
				{ reverse(actual); reverse(next); append(actual, next); segmentUsed[j] = true; }
			}

		if ( (actual.size() > 3) && (dist2(actual.back(), actual.front()) < maxDist) )
			{
			QPoint a1 = actual.takeLast(), a2 = actual.last(), b1 = actual.takeFirst(), b2 = actual.front();
			QPoint p = linesIntersection(a1,a2, b1,b2);
			actual.push_back(p);
			actual.push_front(p);
			}

		result->append(actual);
		}

	if (!segmentUsed[n-1])
		result->append(polylineList.back());

	delete [] segmentUsed;

	return result;
	}*/

/*void qvEllipse(QVPolyline &ellipse, uInt n, float x, float y, float maxRadio, float minRadio, float ang)
	{
	float w = 2*CONST_PI/(n-1);
	float maxArg = (maxRadio+minRadio)/2;
	float minArg = (maxRadio-minRadio)/2;

	for (uInt t = 0; t < n; t++)
		ellipse.append(QPoint (	(uInt) (x + maxArg*cos(ang+w*t) + minArg*cos(ang-w*t)),
						(uInt) (y + maxArg*sin(ang+w*t) + minArg*sin(ang-w*t))
						));
	}

void qvRectangle(QVPolyline &rectangle, int x1, int y1, int x2, int y2)
	{
	rectangle.append(QPoint( x1, y1 ));
	rectangle.append(QPoint( x1, y2 ));
	rectangle.append(QPoint( x2, y2 ));
	rectangle.append(QPoint( x2, y1 ));
	rectangle.append(QPoint( x1, y1 ));
	}

void qvLine(QVPolyline &line, int x1, int y1, int x2, int y2)
	{
	line.append(QPoint( x1, y1 ));
	if (x1 == x2 && y1 == y2)
		return;

	int i,dx,dy,sdx,sdy,dxabs,dyabs,x,y,px,py;
	
	dx=x2-x1;
	dy=y2-y1;
	dxabs=abs(dx);
	dyabs=abs(dy);
	sdx=SIGN(dx);
	sdy=SIGN(dy);
	x=dyabs>>1;
	y=dxabs>>1;
	px=x1;
	py=y1;
	
	if (dxabs>=dyabs)
		for(i=0;i<dxabs;i++)
			{
			y+=dyabs;
			if (y>=dxabs)
				{
				y-=dxabs;
				py+=sdy;
				}
			px+=sdx;
			line.append(QPoint( px, py ));
			}
	else
		for(i=0;i<dyabs;i++)
			{
			x+=dxabs;
			if (x>=dyabs)
				{
				x-=dyabs;
				px+=sdx;
				}
			py+=sdy;
			line.append(QPoint( px, py ));
			}
	}*/
	

///////////////////////////////////////////////////////////////////////////
#define dist2(A,B)	(((A).rx() -(B).rx())*((A).rx() -(B).rx()) + ((A).ry() -(B).ry())*((A).ry() -(B).ry()))
#define dot(A,B)	( (A).rx()*(B).rx() + (A).ry()*(B).ry() )

// simplifyDP():
//  This is the Douglas-Peucker recursive simplification routine
//  It just marks vertices that are part of the simplified polyline
//  for approximating the polyline subchain v[j] to v[k].
//    Input:  tol = approximation tolerance
//            v[] = polyline array of vertex points 
//            j,k = indices for the subchain v[j] to v[k]
//    Output: mk[] = array of markers matching vertex array v[]
/*void simplifyDP( float tol, QPoint* v, int j, int k, int* mk )
	{
	if (k <= j+1) // there is nothing to simplify
		return;

	// check for adequate approximation by segment S from v[j] to v[k]
	int     maxi = j;          // index of vertex farthest from S
	float   maxd2 = 0;         // distance squared of farthest vertex

	QPoint S_P0 = v[j],
		S_P1 = v[k];

	QPoint u = QPoint( S_P1.rx() - S_P0.rx(), S_P1.ry() - S_P0.ry() );
	double  cu = dot(u,u);     // segment length squared
	
	// test each vertex v[i] for max distance from S
	// compute using the Feb 2001 Algorithm's dist_Point_to_Segment()
	// Note: this works in any dimension (2D, 3D, ...)
	QPoint  w;   // vector
	QPoint   Pb;                // base of perpendicular from v[i] to S
	double  b, cw, dv2;        // dv2 = distance v[i] to S squared
	
	for (int i=j+1; i<k; i++)
		{
		// compute distance squared
		w = QPoint( v[i].rx() - S_P0.rx(), v[i].ry() - S_P0.ry() );
		cw = dot(w,u);
		if ( cw <= 0 )   // if perp line doesnt fall on u, it calculates the dist2 to the nearest vertex
			dv2 = dist2(v[i], S_P0);
		else if ( cu <= cw )
			dv2 = dist2(v[i], S_P1);
		else
			{
			// if perp line does fall on u, it calculates the height^2 of the triangle
			b = cw / cu;
			Pb = QPoint( S_P0.rx() + (int) (b * u.rx()), S_P0.ry() + (int)(b*u.ry()) );
			dv2 = dist2(v[i], Pb);
			}
		// test with current max distance squared
		if (dv2 <= maxd2) 
			continue;
		// v[i] is a new max vertex
		maxi = i;
		maxd2 = dv2;
		}
	if (maxd2 > tol * tol)        // error is worse than the tolerance
		{
		// split the polyline at the farthest vertex from S
		mk[maxi] = 1;				// mark v[maxi] for the simplified polyline
							// recursively simplify the two subpolylines at v[maxi]
		simplifyDP( tol, v, j, maxi, mk );	// polyline v[j] to v[maxi]
		simplifyDP( tol, v, maxi, k, mk );	// polyline v[maxi] to v[k]
		}
	// else the approximation is OK, so ignore intermediate vertices
	return;
	}*/

// poly_simplify():
//    Input:  tol = approximation tolerance
//            V[] = polyline array of vertex points 
//            n   = the number of points in V[]
//    Output: sV[]= simplified polyline vertices (max is n)
//    Return: m   = the number of points in sV[]
/*QVPolyline IPE2( float tol, const QVPolyline &V )
	{
	int n=V.size();
	if (n == 0) return V;

	QVPolyline sV;

	QPoint* vt = new QPoint[n];      // vertex buffer
	int*   mk = new int[n];  	// marker buffer
	
	// STAGE 1.  Vertex Reduction within tolerance of prior vertex cluster
	vt[0] = V.at(0);              // start at the beginning

	int i,k,pv;
	for (i=k=1, pv=0; i<n; i++)
		{
		/// \todo ERROR de compilación.
		//if (dist2(V.at(i), V.at(pv)) < tol*tol)
		//	continue;
		vt[k++] = V.at(i);
		pv = i;
		}
	if (pv < n-1)
	vt[k++] = V.at(n-1);      // finish at the end
	
	// STAGE 2.  Douglas-Peucker polyline simplification
	for(int i=0; i<n; ++i)
		mk[i]=0;
	mk[0] = mk[k-1] = 1;       // mark the first and last vertices
	simplifyDP( tol, vt, 0, k-1, mk );
	
	// copy marked vertices to the output simplified polyline
	for (int i=0; i<k; i++)
		{
		if (mk[i])
			sV.append(vt[i]);
			//sV.push_back(vt[i]);
		}
	delete vt;
	delete mk;
	return sV;         // simplyfied polyline
	}*/
}
